#|
Arxiu: LinearAlgebra.lisp.
Contingut: Mtodes d'lgebra linial i matrius.
Autor: Felip Alez Nadal.
ltima actualitzaci: 3-VI-06

|#

#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#


;; Mdul d'lgebra linial per al programa. Aquestes funcions sn necessries per tal de resoldre els sistemes d'equacions linials als quals es redueix el clcul d'splines. 

(require :util)
(provide :ph_LinearAlgebra)

(defpackage :ph_LinearAlgebra 
  (:use #:CL #:UTIL)
  (:nicknames "LinearAlgebra" "LINEARALGEBRA" "LINALG")
  (:export #:INITIALIZE_WITH_ZEROES #:MATRIX_ASIGN_VALUES #:MOSTRA #:MATRIX_IDENTITY #:LU #:PERMUTE #:SOLVE_LINEAR_EQUATIONS #:MATRIX_MULTIPLY #:TRANSPOSE #:INVERSE #:MATRIX_COPY #:SUMA_MATRIUS #:ASIGNA_VALORS_DIAGONAL #:RESTA_MATRIUS)
  (:documentation "This package provides methods for linear algebra.")
)

(in-package :ph_LinearAlgebra)

(defun initialize_with_zeroes(matriu)
  (let ((dimensions nil)(files nil)(columnes nil))
    (setq dimensions (array-dimensions matriu))
    (setq files (car dimensions))
    (setq columnes (car (cdr dimensions)))
    (do ((i 0 (+ i 1)))
	((= i files ) matriu)
	(do ((j 0 (+ j 1)))
	    (( = j columnes) matriu)
	    (setf (aref matriu i j) 0))
	)
    )
  )


(defun suma_matrius (a b)
  "Returns the matrix a + b."
  (let ((dimensions_a nil) (dimensions_b nil) (files_a nil) (files_b nil)(columnes_a nil)(columnes_b nil)(retornable nil)(aux nil))
    (setq dimensions_a (array-dimensions a))
    (setq dimensions_b (array-dimensions b))
    (setq files_a (car dimensions_a))
    (setq columnes_a (car (cdr dimensions_a)))
    (setq files_b (car dimensions_b))
    (setq columnes_b (car (cdr dimensions_b)))
    (unless (= files_a files_b) 
      (print "Impossible sumar dues columnes amb dimensions distintes")
      (return-from suma_matrius nil)
      )
     (unless (= columnes_a columnes_b) 
      (print "Impossible sumar dues columnes amb dimensions distintes")
      (return-from suma_matrius nil)
      )
     (push files_a aux)
     (push columnes_a aux)
     (setq retornable (make-array (list files_a columnes_a)))
     (do ((i 0 (+ i 1)))
	 ((=  i files_a ) retornable)
	 (do ((j 0 (+ j 1)))
	     ((= j columnes_a) t)
	     (setf (aref retornable i j) (+ (aref a i j) (aref b i j)))
	     )
	 )
     )
  )
    

(defun resta_matrius (a b)
  "Returns the matrix a - b."
  (let ((dimensions_a nil) (dimensions_b nil) (files_a nil) (files_b nil)(columnes_a nil)(columnes_b nil)(retornable nil)(aux nil))
    (setq dimensions_a (array-dimensions a))
    (setq dimensions_b (array-dimensions b))
  
    (setq files_a (car dimensions_a))
    (setq columnes_a (car (cdr dimensions_a)))
    (setq files_b (car dimensions_b))
    (setq columnes_b (car (cdr dimensions_b)))
    (unless (= files_a files_b) 
      (print "Impossible sumar dues columnes amb dimensions distintes")
      (return-from resta_matrius nil)
      )
     (unless (= columnes_a columnes_b) 
      (print "Impossible sumar dues columnes amb dimensions distintes")
      (return-from resta_matrius nil)
      )
     (setq retornable (make-array (list files_a columnes_a)))
     (do ((i 0 (+ i 1)))
	 ((=  i files_a ) retornable)
	 (do ((j 0 (+ j 1)))
	     ((= j columnes_a) t)
	     (setf (aref retornable i j) (- (aref a i j) (aref b i j)))
	     )
	 )
     )
  )
 
    ;;Aquesta funci seria millor que tornara la matriu, per tal de poder fer assignacions directes a variables
    (defun matrix_asign_values (matriu valors)
      "Asigns the values of the list valors to matriu, sequentially."
      (let ((dimensions nil)(files nil)(columnes nil))
    (setq dimensions (array-dimensions matriu))
    (setq files (car dimensions))
    (setq columnes (car (cdr dimensions)))
    (do ((i 0 (+ i 1)))
	((= i files) t)
	(do ((j 0 (+ j 1)))
	    ((= j columnes) t)
	    (setf (aref matriu i j) (pop valors))
	    )
	)
    )
  )

(defun show_matrix (matriu) 
  "Shows a matrix."
  (let ((dimensions nil)(files nil)(columnes nil))
    (setq dimensions (array-dimensions matriu))
    (setq files (car dimensions))
    (setq columnes (car (cdr dimensions)))
    (do ((i 0 (+ i 1)))
	((= i files) t)
	(do ((j 0 (+ j 1)))
	    ((= j columnes) t)
	    (format t "~S       " (aref matriu i j))
	    )
	(format t "~%")
	)
    )
  )
(defun matrix_identity (matriu)
  (initialize_with_zeroes matriu)
  (let ((dimensio nil))
    ;;Ac estic suposant que la matriu que ens donen ja s quadrada
    (setq dimensio (car (array-dimensions matriu )))
    (do ((i 0 (+ i 1)))
	((= i dimensio) t)
	( setf (aref matriu i i) 1)
	)
    )
  )
(defun es_quadrada(matriu)
  "Returns T if matriu is a square matrix."
  (let ((aux nil)(dimensions nil))
    (setq dimensions (array-dimensions matriu))
    (if (= (car dimensions ) (car (cdr dimensions ))) (setq aux t)(setq aux nil))
    )
  )


(defun LU ( matriu )  
  "This function makes the LU decomposition of matriu."
  (unless (es_quadrada matriu ) (print "La descomposici LU requereix matrius quadrades")(return-from LU nil))
(let ((N nil)(L nil) (U nil) (P nil)(aux nil)(pivot nil)(max nil)(signum 1)(A nil)(aij nil)(ajj nil)(aik nil)(ajk nil))
  (setq N (car (array-dimensions matriu)))
  (push N aux)
  (push N aux)
  (setq L (make-array aux))
  (setq U (make-array aux))
  (setq P (make-array aux))
  (setq A (make-array aux))
  (matrix_copy matriu A)
;;  (show_matrix A)
  (matrix_identity L)
  (matrix_identity U)
  (matrix_identity P)
  (do ((j 0 (+ j 1)))
      ((= j N))
      (setq max (aref A j j))
      (setq pivot j)
      ;;Abans que res busque el pivot
      (do ((i (+ j 1) (+ i 1)))
	  ((= i N) t)
	  (when (> (aref A i j) max)
	    (setq max (aref A i j))
	    (setq pivot i)
	    )
	  )
      (unless (= j pivot)
	(permute A j pivot)
	(permute P j pivot)
	(setq signum (- signum))
	))
  (do ((j 0 (+ j 1)))
      ((= j N) t)
      (setq ajj (aref A j j))
      (do ((i (+ j 1) (+ i 1)))
	  ((= i N) t)
	  (setq aij (/(aref A i j) ajj))
	  (setf (aref A i j) aij);
	  (do ((k (+ j 1) (+ k 1)))
	      ((= k N) t)
	      (setq aik (aref A i k))
	      (setq ajk (aref A j k))
	      (setf (aref A i k) (- aik (* aij ajk)))
	      )
	  )
      )
  (do ((j 0 (+ j 1)))
      ((= j N) t)
      (do ((i 0 (+ i 1)))
	  ((= i N) t)
	  (when (> i j) ;; Hem d'assignar a L i j
	    (setf (aref L i j) (aref A i j))
	    )
	  (unless (> i j) ;; Hem d'assignar a Uij
	    (setf (aref U i j)(aref A i j))
	    )
	  )
      )
  (list L U P signum)
  )
)

;;ULL!! Comprovar si i i j existeixen!
(defun permute (matriu i j)
  (let ((auxiliar nil)(columnes nil)(dimensions nil))
    (setq dimensions (array-dimensions matriu))
    (setq columnes (car (cdr dimensions)))
    (do ((k 0 (+ k 1)))
	((= k columnes) t)
	(setq auxiliar (aref matriu i k))
	(setf (aref matriu i k) (aref matriu j k))
	(setf (aref matriu j k) auxiliar)
	
	)
    )
  )


(defun solve_linear_equations ( A B )
  #| A s la matriu del sistema
     B s el vector de termes independents !!! Ha d'estar en la forma correcta!
     X s un vector on es dipositaran els resultats
  
     He de resoldre el segent sistema:
                  AX = B
     Transformant-lo en :
                  PAX = LUX = PB
     Ho far resolent :
                  LY=PB = W
     I a continuaci :
                  UX = Y
     Aquests dos sistemes sn fcils de resoldre, el primer per substituci progressiva i el segon per substituci regressiva.

  |#

;;Falta comprovar si A s quadrada i si B i X tenen les dimensions adequades!
  (let ((res_LU nil) (L nil) (U nil) (signum nil)(P nil)(B_per nil)(Y  nil)(dimensions nil) (N nil)(aux nil)(X nil))
    ;;Primerament faig una descomposici LU i deixe els resultats en la llista res_LU
    (setq res_LU (LU A))
    (setq L (pop res_LU))
    (setq U (pop res_LU))
    (setq P (pop res_LU))
    (setq signum (pop res_LU))
    ;;Ara aplique la permuteci a B
    (setq B_per (matrix_multiply P B))
    (setq dimensions (array-dimensions A))
    (setq N (car dimensions))
    (push 1 aux)
    (push N aux)
    (setq Y (make-array aux))
    (setq X (make-array aux))
    (initialize_with_zeroes Y)
    ;; Ara ja puc resoldre el primer sistema d'equacions. Per primer necessite conixer les dimensions de la matriu A
    (do ((i 0 (+ i 1)))
	((= i N) t)
	(let ((suma 0))
	  (do (( j 0 (+ j 1)))
	      ((= j i) t)
	      (setq suma (+ suma (* (aref L i j) (aref Y j 0))))
	      )
	  (setf (aref Y i 0) (-(aref B_per i 0) suma))
	  )
	)
    (do ((i (- N 1) (- i 1)))
	((= i -1) t)
	(let (( suma 0))
	  (do (( j (+ i 1) (+ j 1)))
	      ((= j N) t)
	      (setq suma (+ suma (* (aref U i j) (aref X j 0))))
	      )
	  (setf (aref X i 0) (/(- (aref Y i 0) suma) (aref U i i)))
	  )
	)
    ;;Les dues lnies segents sobren, per com que de moment estic en proves les deixe per a fer-ho rpid
;;    (format t "Vector X ~%")
;;    (show_matrix X)
    X
    )
  
)
(defun matrix_multiply (A B)
;;Anem per feina
  (let ((dimensions_A nil)(dimensions_B nil)(files_A nil)(files_B nil)(columnes_A nil)(columnes_B nil)(resultat nil))
    (setq dimensions_A (array-dimensions A))
    (setq dimensions_B (array-dimensions B))
    (setq files_A (car dimensions_A))
    (setq columnes_A (car (cdr dimensions_A)))
    (setq files_B (car dimensions_B))
    (setq columnes_B (car (cdr dimensions_B)))
    (if (eq files_A nil) 
	(setq files_A 0)
      )
    
    (if (eq files_B nil) 
	(setq files_B 0)
      )
  
    (if (eq columnes_A nil) 
	(setq columnes_A 0)
      )
    
    (if (eq columnes_B nil)
	(setq columnes_B 0)
      )
    
    ;;Comprove si es poden multiplicar
    (unless (= columnes_A files_B) 
      (format t "Error!!! No s possible multiplicar una matriu de (~D , ~D) per una de (~D, ~D)~%" files_A columnes_A files_B columnes_B)
      (return-from matrix_multiply nil)
      )
    (let ((aux nil))
      (push columnes_B aux)
      (push files_A aux)
      (setq resultat (make-array  aux))
      )
    (do ((i 0 (+ i 1)))
      ((= i files_A) t)
      (do ((j 0 (+ j 1)))
	  ((= j columnes_B) t)
	  (let ((suma 0))
	    (do ((k 0  (+ k 1)))
		((= k columnes_A) t)
		(setq suma (+ suma (* (aref A i k) (aref B k j))))
		;;		(format t "Suma val ~S~%" suma)
		)
	    ;;	    (format t "Assignant element ~S, ~S" i j)
	    (setf (aref resultat i j) suma)
	    )
	  )
      )
    resultat
    )
  )


(defun matrix_copy (font desti)
  (let ((dimensions_font nil)(dimensions_desti nil)(files_f nil)(files_d nil)(columnes_f nil)(columnes_d nil))
    (setq dimensions_font (array-dimensions font))
    (setq dimensions_desti (array-dimensions desti))
    (setq files_f (car dimensions_font))
    (setq files_d (car dimensions_desti))
    (setq columnes_f (car (cdr dimensions_font)))
    (setq columnes_d (car (cdr dimensions_desti)))
    (unless (= files_f files_d) (print "Les matrius no tenen el mateix nombre de files, per tant no s possible copiar-les!!!")(return-from matrix_copy nil))
    (unless (= columnes_f columnes_d ) (print "Les matrius no tenen el mateix nombre de columnes, per tant no s possible copiar-les")(return-from matrix_copy nil))
    ;;Ac falta alguna comprovaci per sortir
    (do ((i 0 (+ i 1)))
	((= i files_f) t)
	(do ((j 0 (+ j 1)))
	    ((= j columnes_f) t)
	    (setf (aref desti i j) (aref font i j))
	    )
	)
    )
  
  )


(defun transpose (A)
  "Returns the transposed matrix of A"
  (let ((files 0) (columnes 0)(B nil)(auxiliar nil))
    (setq auxiliar (array-dimensions A))
    (setq files (car auxiliar))
    (setq columnes (car (cdr auxiliar)))
    (setq auxiliar (list columnes files))
    (setq B (make-array auxiliar))
    (do ((i 0 (+ i 1)))
	((= i files) t)
      (do ((j 0 (+ j 1)))
	  ((= j columnes ) t)
	(setf (aref B j i) (aref A i j))
	)
      )
    B
    )
  )

(defun columna (matriu nombre)
  ;;Torna la columna nombre de la matriu
  (let ((columna nil)(dimensions nil)(files 0)(aux (list 1)))
    (setq dimensions (array-dimensions matriu))
    (setq files (car dimensions))
    (push files aux)
    (setq columna (make-array aux))
    ;;Una vegada creada la columna amb les dimensions adequades, simplement la rellene
    (do ((i 0 (+ i 1)))
	((= i files) columna)
	(setf (aref columna i 0) (aref matriu i nombre))
	)
    )
  )


(defun columnes (matriu)
  ;;Torna una llista amb totes les columnes que formen matriu
  (let ((dimensions nil) (columnes 0) (aux nil))
    (setq dimensions (array-dimensions matriu))
    (setq columnes (car (cdr dimensions)))
    (setq aux nil)
    (do ((i 0 (+ i 1)))
	((= i columnes)(reverse aux)) ;;Torna la llista, convenientment invertida, en finalitzar el bucle
	(push (columna matriu i) aux)
	)
    )
  )


(defun matriu_desde_columnes (llista_columnes)
  ;;Aquest mtode torna una matriu construda a partir d'una llista de columnes. Ull!! Les columnes tenen totes dimensions n x 1
  (let ((files 0)(dimensions nil) (columnes 0)(aux nil)(matriu))
    ;;Ull!! Falta comprovar que la llista no siga nil
    (setq columnes (length llista_columnes))
    ;;Pose a aux la primera columna
    (setq aux (pop llista_columnes))
    (setq dimensions (array-dimensions aux))
    ;;Torne a posar-la dins per coherncia
    (push aux llista_columnes)
    (setq files (car dimensions))
    (setq aux nil)
    (push columnes aux)
    (push files aux)
    ;;Construisc la matriu
    (setq matriu (make-array aux))
    
    (do ((i 0 (+ i 1)))
	((= i files) matriu)
	(setq aux (pop llista_columnes))
	(do ((j 0 (+ j 1)))
	    ((= j columnes) t)
	    (setf (aref matriu j i) (aref aux j 0))
;;	    (format t "Element i,j ~S,~S s ~S ~%" i j (aref aux j 0))
	    )
	)
    )
  )


(defun inverse (matriu)
  "Returns the inverse of matriu."
  (let ((aux nil) (dimensions nil)(files 0)(columnes 0)(I nil)(columnes_I nil))
    ;; Conseguisc la llista amb les columnes de la matriu original
    (setq dimensions (array-dimensions matriu))
    (setq columnes (car (cdr dimensions)))
    (setq files (car dimensions))
    (unless (= files columnes) (print "Noms les matrius quadrades tenen inverse"))
    (setq aux nil)
;;    (defvar aux nil)
    (push  columnes aux)
    (push files aux)
    ;;A posteriori cree una matriu quadrada i la transforme en la identitat
    (setq I (make-array aux))
    (matrix_identity I)
    (setq columnes_I (columnes I))
    (setq aux (mapcar #'solve_linear_equations (repeat matriu columnes) columnes_I ))
    (matriu_desde_columnes aux)
    )
  )

(defun asigna_valors_columna (matriu columna n_columna)
  ;;Assigna columna com a valors de la columna n_columna de matriu
  (let ((dimensions nil)(files nil)(columnes nil))
    (setq dimensions (array-dimensions matriu))
    (setq files (car dimensions))
    (setq columnes (car (cdr dimensions)))
    (do ((i 0 (+ i 1)))
	((= i columnes) t)
	(setf (aref matriu i n_columna ) (aref columna i 0))
	)
    )
  )


(defun asigna_valors_diagonal (matriu valors)
  (let* ((dimensions (array-dimensions matriu))(files (car dimensions)))
    (unless (es_quadrada matriu) 
      (print "Nomes les matrius quadrades tenen diagonal")
      (return-from asigna_valors_diagonal nil)
      )
    (do ((i 0 (+ i 1)))
	((= i files) t)
      (setf (aref matriu i i) (pop valors))
      )
    )
  t
  )
