#|
Arxiu: minquad.lisp.
Contingut: Implementacions de mtodes d'ajust per mnims quadrats.
Autor: Felip Alez Nadal.
ltima actualitzaci: 4-VI-06

|#


#|BSD License

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

   1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
   2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE. |#



;;NOTA!!! He de comprovar si el mdul est carregat abans de carregar-lo!!!

(provide :ph_MinQuad)
;(require :LinearAlgebra "LinearAlgebra.lisp")
;(require :estadistica "estadistica.lisp")

(defpackage :ph_MinQuad
  (:use :ph_LinearAlgebra :CL  :ph_estatistical :util)
  (:nicknames "MINQUAD" "AJUSTS")
  (:export #:MIN_QUAD_RECTA #:MIN_QUAD_RECTA_SENSE_PES #:MIN_QUAD_LINIAL #:MIN_QUAD_LINIAL_SENSE_PES)
  (:documentation "This package provides functions to adjust experimental data to linear models.")
  )

(in-package :ph_MinQuad)

(defun min_quad_recta (llista_x llista_y llista_errors) 
"Calcula l'ajust a una recta d'un conjunt de dades. Tots els seus arguments sn llistes. Torna una llista en el format ( (a b) C), on a s l'ordenada en l'origen, b el pendent i C la matriu de covariana"
  (let ((files nil)(_1_ 0)(_y_ 0)(_x_ 0)(_xx_ 0)(_xy_ 0)(_yy_ 0)(_delta_ 0)(X nil)(A_B nil)(C nil)(A nil)(B nil))
    (setq files(length llista_x ))
    (setq _1_ (apply #'+ (mapcar #'(lambda (x) (/ 1 (expt x 2))) llista_errors)))
    (setq _x_ (apply #'+ (mapcar #'(lambda (x y) (/ x  (expt y 2))) llista_x llista_errors))) 
    (setq _y_ (apply #'+ (mapcar #'(lambda (x y) (/ x  (expt y 2))) llista_y llista_errors)))
    (setq _xx_ (apply #'+ (mapcar #'(lambda (x y) (/ (expt x 2) (expt y 2))) llista_x llista_errors)))
    (setq _yy_ (apply #'+ (mapcar #'(lambda (x y) (/ (expt x 2) (expt y 2))) llista_y llista_errors)))
    (setq _xy_ (apply #'+ (mapcar #'(lambda (x y z) (/ (* x y) (expt z 2))) llista_x llista_y llista_errors)))
    ;;Una vegada calculades les quantitats, calcule la matriu del sistema
    (setq A (make-array '(2 2)))
    (setq B (make-array '(2 1)))
    (setq X (make-array '(2 1)))
    (matrix_asign_values A (list _1_ _x_ _x_ _xx_))
    (matrix_asign_values B (list _y_ _xy_))
    ;;(mostra A)
    ;;(mostra B)
    (setq A_B (solve_linear_equations A B ))
    ;;Ara vaig a per la matriu de covariana
    (setq _delta_  (- (* _xx_ _1_) (*  _x_ _x_)))
    (setq C (make-array '(2 2)))
    (matrix_asign_values C (list (/ _xx_ _delta_) (/ (- _x_ ) _delta_) (/ (- _x_) _delta_) (/ _1_  _delta_)))
    ;;Torne una llista amb A_B i C
    (list A_B C)
    )
  )

(defun min_quad_recta_sense_pes (llista_x llista_y)
"Calcula l'ajust a una recta d'un conjunt de dades. Assigna com a error de cada mesura la desviaci estndar."
	(let ((llista_errors nil))
	  ( setq llista_errors (repeat (desviacio_estandar llista_y )(length llista_y)))
	  ( min_quad_recta llista_x llista_y llista_errors)
	  )
	)
(defun min_quad_linial (llista_x llista_y llista_errors llista_funcions)
  "Adjusts data to a linial model"
  (let ((aux nil)(files (length llista_x))(n_funcions (length llista_funcions))(A nil) (V nil)(y nil)(A_t nil) (V_1 nil)(teta nil)(C nil)(C_1 nil))
    
    ;;n = 0
    (setq aux (list files n_funcions))
    (setq A (make-array aux))
    (setq aux (list files files))
    (setq V (make-array aux))
    (setq aux (list files 1))
    (setq y (make-array aux))
    ;;A = Matrix([[0.0]*n_funcions]*files)
    ;;V =array( [[0.0]*files]*files)
    ;;y = Matrix([[0.0]]*files)
    (asigna_valors_diagonal V (mapcar #'(lambda (x) (expt x 2)) llista_errors))
    ;;asigna_valors_diagonal (V, map (lambda x : x**2 , llista_errors))
    ;;#Ja tinc V en condicions
    ;;while n < files :
    (do ((n 0 (+ n 1)))
	((= n files) t)
      (setf (aref y n 0) (pop llista_y))
      ;;      y[n,0] = llista_y[n]
      (do ((j 0 (+ j 1)))
	  ((= j n_funcions) t)
        ;;while j < n_funcions :
	(setf (aref A n j) (funcall (nth j llista_funcions) (nth n llista_x)))
	;;            A[n,j] = llista_funcions[j] (llista_x[n])
	;;          j += 1
	;;    n+=1
	)
      )
    (setq A_t (transpose A))
    (setq V_1 (inverse V))
    ;;A_t = transposa(A)
    ;;V_1 = inverse(V)
    (setq C_1 (matrix_multiply A_t (matrix_multiply V_1 A)))
    ;;C_1 = A_t*V_1*A
    (setq C (inverse C_1))
    ;;C = inverse(C_1)
    ;;#Ok!! Ja tinc C, A_t i Y. Ara vaig a resoldre el sistema
    (setq teta (matrix_multiply C (matrix_multiply A_t (matrix_multiply V_1 y))))
    ;;teta = C*A_t*V_1*y
    ;;#teta es el vector de parametres, C la matriu de covariansa
    ;;return teta,C
    (values teta C)
    )
  )

(defun min_quad_linial_sense_pes (llista_x llista_y llista_funcions)
  "Calcula l'ajust a una recta d'un conjunt de dades. Assigna com a error de cada mesura la desviaci estndar."
  (let ((llista_errors nil))
    ( setq llista_errors (repeat (desviacio_estandar llista_y )(length llista_y)))
    ( min_quad_linial llista_x llista_y llista_errors llista_funcions)
    )
  )
